/*
 * usersfile.c - implementation of UsersFile based HOTP validation
 * Copyright (C) 2009-2024 Simon Josefsson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <config.h>
#undef GNULIB_POSIXCHECK	/* too many complaints for now */

#include "oath.h"

#include <stdio.h>		/* For snprintf, getline. */
#include <stdlib.h>		/* For free. */
#include <unistd.h>		/* For ssize_t. */
#include <fcntl.h>		/* For fcntl. */
#include <errno.h>		/* For errno. */
#include <limits.h>		/* For PATH_MAX & friends. */
#include <sys/stat.h>		/* For S_IRUSR, S_IWUSR. */
#include <sys/wait.h>		/* For wait */
#include <sys/stat.h>		/* For stat */
#include <grp.h>		/* For setgroups */

#ifndef _WIN32

struct usersfile_ctx {
    const char *path;
    const char *basename;  /* basename of path, points into `path` */
    int parent_fd;         /* file descriptor for the parent directory of the usersfile */
    int fd;                /* file descriptor for the usersfile */
    struct stat st;        /* stat information for the usersfile */
};

/*
 * Upgrade a file descriptor opened with O_PATH to a fully functional file
 * descriptor.
 *
 * To achieve this the file is reopened via /proc, which is supported by the
 * Linux kernel. `fd` needs to point to the currently open file descriptor. On
 * success it will be replaced by the new upgraded file descriptor, while the
 * original file descriptor will be closed.
 *
 * `flags` are passed to `open()` for the new file descriptor.
 */
static int
reopen_path_fd (int *fd, int flags)
{
  /* we need to open /proc/self/fd/<int>, so the path won't get too long here */
  char proc_path[128];
  int res = snprintf(proc_path, sizeof(proc_path), "/proc/self/fd/%d", *fd);

  if (res < 0 || res >= sizeof(proc_path))
      return OATH_PRINTF_ERROR;

  int newfd = open(proc_path, flags);

  if (newfd < 0)
      return OATH_FILE_OPEN_ERROR;

  close(*fd);
  *fd = newfd;
  return OATH_OK;
}

static void
init_usersfile_ctx(struct usersfile_ctx *ctx, const char *path)
{
  ctx->path = path;
  ctx->basename = NULL;
  ctx->parent_fd = -1;
  ctx->fd = -1;
  memset(&ctx->st, 0, sizeof(ctx->st));
}

static void
destroy_usersfile_ctx(struct usersfile_ctx *ctx)
{
   if (ctx->parent_fd != -1)
     {
         close (ctx->parent_fd);
         ctx->parent_fd = -1;
     }

   if (ctx->fd != -1)
     {
         close (ctx->fd);
         ctx->fd = -1;
     }

   /* reset everything but keep the path so it might be reused */
   init_usersfile_ctx(ctx, ctx->path);
}

/*
 * Obtain a lock for the usersfile. The lock is placed on the usersfile itself
 * as found in `ctx->fd`
 *
 * On success the lock on `ctx->fd` has been correctly obtained.
 */
static int
lock_usersfile (struct usersfile_ctx *ctx)
{
  /*
   * There exist three file locking APIs:
   *
   * - flock(): this would be the simplest API, but it doesn't properly support
   *   network file systems like NFS, which then causes a transparent fallback
   *   to fcntl() file locking.
   * - fcntl using F_SETLCK & friends: this lock is not based on the open file
   *   description and thus cannot be inherited to child processes, which we
   *   need to do.
   * - fcntl using F_OFD_SETLCK & friends: this is a Linux specific lock that
   *   _is_ based on the open file description. It seems like the best bet for
   *   our scenario.
   *
   * Since we are potentially running in PAM module context, we have to
   * take a local DoS scenario into account here, where the unprivileged user
   * holds the lock, preventing us from ever getting it.
   *
   * There's no file locking API supporting a timeout (except for using a
   * SIGALRM timer to interrupt the system call). Using asynchronous signals
   * in a library is not so great. Thus make a best effort polling attempt:
   *
   * `F_OFD_SETLK` polls for the lock. If we cannot get it, sleep half a
   * second and retry. Do this for at max 15 seconds, else fail.
   */

  struct flock fl;
  memset(&fl, 0, sizeof(fl));
  /* lock the entire file with a write lock */
  fl.l_type = F_WRLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;

  for (int i = 0; i < 30; i++) {
      if (fcntl(ctx->fd, F_OFD_SETLK, &fl) == 0)
          return OATH_OK;

      if (errno == EACCES || errno == EAGAIN)
        usleep(1000 * 500);
      else
        break;
  }

  return OATH_FILE_LOCK_ERROR;
}

/*
 * After traversing all directory path elements this function actually opens
 * the target usersfile. `ctx->parent_fd` must be valid.
 *
 * This function takes care of the locking logic, which is a bit complicated,
 * since we use the usersfile itself for locking. This is done, because we
 * don't want to clutter arbitrary directories with lockfiles, possibly making
 * the locking also less robust (e.g. if users delete them interactively).
 *
 * Since we don't actually write to the usersfile, but replace it atomically,
 * to prevent any inconsistent state to ever be stored to disk, we need a
 * recovery mechanism if we obtain a lock on the file, but the file has
 * already been replaced by a new version. This situation is detected by
 * opening the file again after the lock has been placed and comparing the
 * inode numbers. If they no longer match, then the new file has to be locked
 * instead.
 *
 * On successful return ctx->fd will be valid and locked and ctx->st will
 * contain the current stat information for the usersfile.
 */
static int
finish_open_usersfile (struct usersfile_ctx *ctx)
{
  const int oflags = O_RDONLY|O_PATH|O_CLOEXEC|O_NOFOLLOW;
  ctx->fd = openat(ctx->parent_fd, ctx->basename, oflags);

  if (ctx->fd < 0)
    return errno == ENOENT ? OATH_NO_SUCH_FILE : OATH_FILE_OPEN_ERROR;

  if (fstat(ctx->fd, &ctx->st) != 0)
      return OATH_FILE_STAT_ERROR;

  /* don't allow hard-linked files, which would allow to fool our logic -
   * this can only happen if protected_hardlinks is disabled in the kernel,
   * though */
  if (ctx->st.st_nlink > 1)
    {
      close(ctx->fd);
      return OATH_FILE_OPEN_ERROR;
    }

  /* lock and retry opening until all is consistent, abort after a couple of
   * times, it's unlikely that we race all the time (could be a DoS attempt) */
  for (int i = 0; i < 5; i++)
    {
      /* deny world-writable or special usersfile */
      if ((ctx->st.st_mode & S_IWOTH) != 0 || !S_ISREG(ctx->st.st_mode))
          return OATH_FILE_OPEN_ERROR;

      /* we need to open it read-write for write-locking it via fcntl(),
       * otherwise we wouldn't need write access for the file, since we'll
       * atomically replace it with a new one. */
      int err = reopen_path_fd(&ctx->fd, O_RDWR|O_CLOEXEC|O_NOCTTY);
      if (err != OATH_OK)
          return err;

      err = lock_usersfile(ctx);
      if (err != OATH_OK)
          return err;

      /*
       * we now own a lock on the usersfile, but another process might already
       * have replaced the file in question by new version. Thus we need to
       * check whether the file is still there and is the same as the one we
       * have opened. Otherwise a race occurred an we need to retry.
       */
      int check_fd = openat(ctx->parent_fd, ctx->basename, oflags);
      struct stat check_st;
      err = fstat(check_fd, &check_st);
      if (err != OATH_OK)
        {
          close(check_fd);
          return err;
        }

      /* comparing the inode should be enough, since parent_fd didn't change,
       * so it should be the same file system */
      if (ctx->st.st_ino != check_st.st_ino)
        {
          /* race occurred, retry using the new FD */
          close(ctx->fd);
          ctx->fd = check_fd;
          memcpy(&ctx->st, &check_st, sizeof(ctx->st));
          continue;
        }

      /* we own the lock and the file is still in place, we did it */
      close(check_fd);

      /* now also reopen the parent directory FD, so it can be used for
       * fsync() later on. */
      err = reopen_path_fd(&ctx->parent_fd, O_RDONLY|O_CLOEXEC|O_DIRECTORY);
      if (err != OATH_OK)
          return err;

      return OATH_OK;
    }

  /* maximum number of locking attempts exceeded */
  return OATH_FILE_LOCK_ERROR;
}

static int
parse_type (const char *str, unsigned *digits, unsigned *totpstepsize)
{
  *totpstepsize = 0;
  if (strcmp (str, "HOTP/E/6") == 0
      || strcmp (str, "HOTP/E") == 0 || strcmp (str, "HOTP") == 0)
    *digits = 6;
  else if (strcmp (str, "HOTP/E/7") == 0)
    *digits = 7;
  else if (strcmp (str, "HOTP/E/8") == 0)
    *digits = 8;
  else if (strncmp (str, "HOTP/T30", 8) == 0)
    {
      *totpstepsize = 30;
      if (strcmp (str, "HOTP/T30") == 0 || strcmp (str, "HOTP/T30/6") == 0)
	*digits = 6;
      else if (strcmp (str, "HOTP/T30/7") == 0)
	*digits = 7;
      else if (strcmp (str, "HOTP/T30/8") == 0)
	*digits = 8;
      else
	return -1;
    }
  else if (strncmp (str, "HOTP/T60", 8) == 0)
    {
      *totpstepsize = 60;
      if (strcmp (str, "HOTP/T60") == 0 || strcmp (str, "HOTP/T60/6") == 0)
	*digits = 6;
      else if (strcmp (str, "HOTP/T60/7") == 0)
	*digits = 7;
      else if (strcmp (str, "HOTP/T60/8") == 0)
	*digits = 8;
      else
	return -1;
    }
  else
    return -1;

  return 0;
}

static const char *whitespace = " \t\r\n";
# define TIME_FORMAT_STRING "%Y-%m-%dT%H:%M:%SL"

static int
parse_usersfile (const char *username,
		 const char *otp,
		 size_t window,
		 const char *passwd,
		 time_t *last_otp,
		 FILE *infh,
		 char **lineptr, size_t *n, uint64_t *new_moving_factor,
		 size_t *skipped_users)
{
  int bad_password = 0;

  *skipped_users = 0;

  while (getline (lineptr, n, infh) != -1)
    {
      char *saveptr;
      char *p = strtok_r (*lineptr, whitespace, &saveptr);
      unsigned digits, totpstepsize;
      char secret[32];
      size_t secret_length = sizeof (secret);
      uint64_t start_moving_factor = 0;
      int rc = 0;
      char *prev_otp = NULL;

      if (p == NULL)
	continue;

      /* Read token type */
      if (parse_type (p, &digits, &totpstepsize) != 0)
	continue;

      /* Read username */
      p = strtok_r (NULL, whitespace, &saveptr);
      if (p == NULL || strcmp (p, username) != 0)
	continue;

      /* Read password. */
      p = strtok_r (NULL, whitespace, &saveptr);
      if (passwd)
	{
	  if (p == NULL)
	    continue;
	  if (strcmp (p, "-") == 0)
	    {
	      if (*passwd != '\0')
		{
		  bad_password = 1;
		  rc = OATH_BAD_PASSWORD;
		}
	    }
	  else if (strcmp (p, "+") == 0)
	    {
	      /* Externally verified. */
	    }
	  else if (strcmp (p, passwd) != 0)
	    {
	      bad_password = 1;
	      rc = OATH_BAD_PASSWORD;
	    }
	  if (rc == OATH_BAD_PASSWORD)
	    {
	      (*skipped_users)++;
	      continue;
	    }
	  bad_password = 0;
	}

      /* Read key. */
      p = strtok_r (NULL, whitespace, &saveptr);
      if (p == NULL)
	continue;
      rc = oath_hex2bin (p, secret, &secret_length);
      if (rc != OATH_OK)
	return rc;

      /* Read (optional) moving factor. */
      p = strtok_r (NULL, whitespace, &saveptr);
      if (p && *p)
	{
	  char *endptr;
	  unsigned long long int ull = strtoull (p, &endptr, 10);
	  if (endptr && *endptr != '\0')
	    return OATH_INVALID_COUNTER;
	  start_moving_factor = ull;
	}

      /* Read (optional) last OTP */
      prev_otp = strtok_r (NULL, whitespace, &saveptr);

      /* Read (optional) last_otp */
      p = strtok_r (NULL, whitespace, &saveptr);
      if (p)
	{
	  struct tm tm;
	  char *ts;

	  ts = strptime (p, TIME_FORMAT_STRING, &tm);
	  if (ts == NULL || *ts != '\0')
	    return OATH_INVALID_TIMESTAMP;
	  tm.tm_isdst = -1;
	  if (last_otp)
	    {
	      *last_otp = mktime (&tm);
	      if (*last_otp == (time_t) - 1)
		return OATH_INVALID_TIMESTAMP;
	    }
	}

      if (prev_otp && strcmp (prev_otp, otp) == 0)
	return OATH_REPLAYED_OTP;

      if (totpstepsize == 0)
	rc = oath_hotp_validate (secret, secret_length,
				 start_moving_factor, window, otp);
      else if (prev_otp)
	{
	  int prev_otp_pos, this_otp_pos, tmprc;
	  rc = oath_totp_validate2 (secret, secret_length,
				    time (NULL), totpstepsize, 0, window,
				    &this_otp_pos, otp);
	  if (rc == OATH_INVALID_OTP)
	    {
	      (*skipped_users)++;
	      continue;
	    }
	  if (rc < 0)
	    return rc;
	  tmprc = oath_totp_validate2 (secret, secret_length,
				       time (NULL), totpstepsize, 0, window,
				       &prev_otp_pos, prev_otp);
	  if (tmprc >= 0 && prev_otp_pos >= this_otp_pos)
	    return OATH_REPLAYED_OTP;
	}
      else
	rc = oath_totp_validate (secret, secret_length,
				 time (NULL), totpstepsize, 0, window, otp);
      if (rc == OATH_INVALID_OTP)
	{
	  (*skipped_users)++;
	  continue;
	}
      if (rc < 0)
	return rc;
      *new_moving_factor = start_moving_factor + rc;
      return OATH_OK;
    }

  if (*skipped_users)
    {
      if (bad_password)
	return OATH_BAD_PASSWORD;
      else
	return OATH_INVALID_OTP;
    }

  return OATH_UNKNOWN_USER;
}

static int
update_usersfile2 (const char *username,
		   const char *otp,
		   FILE *infh,
		   FILE *outfh,
		   char **lineptr,
		   size_t *n, char *timestamp, uint64_t new_moving_factor,
		   size_t skipped_users)
{
  size_t got_users = 0;

  while (getline (lineptr, n, infh) != -1)
    {
      char *saveptr;
      char *origline;
      const char *user, *type, *passwd, *secret;
      int r;
      unsigned digits, totpstepsize;

      origline = strdup (*lineptr);

      type = strtok_r (*lineptr, whitespace, &saveptr);
      if (type == NULL)
	goto skip_line;

      /* Read token type */
      if (parse_type (type, &digits, &totpstepsize) != 0)
	goto skip_line;

      /* Read username */
      user = strtok_r (NULL, whitespace, &saveptr);
      if (user == NULL || strcmp (user, username) != 0
	  || got_users++ != skipped_users)
	goto skip_line;

      passwd = strtok_r (NULL, whitespace, &saveptr);
      if (passwd == NULL)
	passwd = "-";

      secret = strtok_r (NULL, whitespace, &saveptr);
      if (secret == NULL)
	secret = "-";

      r = fprintf (outfh, "%s\t%s\t%s\t%s\t%llu\t%s\t%s\n",
		   type, username, passwd, secret,
		   (unsigned long long) new_moving_factor, otp, timestamp);
      free (origline);
      if (r <= 0)
	return OATH_PRINTF_ERROR;
      continue;

    skip_line:
      r = fprintf (outfh, "%s", origline);
      free (origline);
      if (r <= 0)
	return OATH_PRINTF_ERROR;
      continue;
    }

  return OATH_OK;
}

/*
 * create a new file in the directory referred to by ctx->parent_fd. A unique
 * filename will be selected and written out to `newname`.
 */
static int
create_new_usersfile(struct usersfile_ctx *ctx, char *newname)
{
  int err = OATH_OK;
  newname[0] = '\0';

  /* create an unnamed temporary file, this way we can fix the file mode
     without anybody else being able to access the file */
  int fd = openat(ctx->parent_fd, ".", O_TMPFILE|O_WRONLY|O_CLOEXEC, 0600);
  if (fd < 0)
    return OATH_FILE_OPEN_ERROR;

  /* make sure the mode is as we want it, since umask might have changed the outcome. */
  if (fchmod(fd, 0600) != 0)
    {
        err = OATH_FILE_CHOWN_ERROR;
        goto out;
    }

  /* there's nothing like mkostmpat() where we can use our parent_fd.
   * tmpname() & friends are deprecated and also not fully suitable here.
   *
   * what we're actually missing here is an additional flag LINKAT_REPLACE
   * which would allow to atomically replace the original file, instead of
   * using renameat(). This doesn't exist yet, though.
   *
   * linkat() doesn't follow symlinks or overwrite files, so we're safe here
   * against any shenanigans. The user owning parent_fd can try to guess the
   * filename we're using here and thus DoS us. Setup an arbitrary limit of
   * creation attempts to prevent an infinite loop in such situations. Such a
   * bad actor would then only DoS itself, preventing login.
   *
   * Shared world-writable directories should never be used for the usersfile,
   * this would be a configuration error, thus we don't try to protect against
   * such scenarios.
   *
   * An alternative would be using rand(), but then we'd need to also seed it,
   * with possible process wide side effects, which is also not great.
   */

  int ret = snprintf(newname, NAME_MAX, "%s.new.%d", ctx->basename, getpid());
  if (ret < 0 || ret >= NAME_MAX)
    {
      err = OATH_PRINTF_ERROR;
      goto out;
    }

  /* we need to specify /proc/self/fd/<int>, so the path won't get too long here */
  char proc_path[128];
  ret = snprintf(proc_path, sizeof(proc_path), "/proc/self/fd/%d", fd);
  if (ret < 0 || ret >= NAME_MAX)
    {
      err = OATH_PRINTF_ERROR;
      goto out;
    }

  /* we cannot reliably use AT_EMPTY_PATH here, since it can require the
   * CAP_DAC_READ_SEARCH capability when running as non-root. Starting with
   * kernel 6.10 this requirement has been softened, but we need to stay
   * backward compatible. Linking the magic link in /proc into the directory
   * works without extra capabilities.
   * For this workaround to function AT_SYMLINK_FOLLOW _must_ be specified
   * so this is a conscious decision.
   */
  if (linkat(AT_FDCWD, proc_path, ctx->parent_fd, newname, AT_SYMLINK_FOLLOW))
    {
      err = OATH_FILE_CREATE_ERROR;
    }

out:
  if (err != OATH_OK)
    {
        if (fd >= 0)
            close(fd);
        return err;
    }

  return fd;
}

static int
update_usersfile (struct usersfile_ctx *ctx,
		  const char *username,
		  const char *otp,
		  FILE *infh,
		  char **lineptr,
		  size_t *n, char *timestamp, uint64_t new_moving_factor,
		  size_t skipped_users)
{
  int rc;

  /* Rewind input file. */
  {
    int pos;

    pos = fseeko (infh, 0L, SEEK_SET);
    if (pos == -1)
      return OATH_FILE_SEEK_ERROR;
    clearerr (infh);
  }

  char newfilename[NAME_MAX];

  /* Open the "new" file. We aim for atomic replacement of the old file to
   * address possible power failure or system lockup scenarios. */
  int outfd = create_new_usersfile(ctx, newfilename);
  if (outfd < 0)
    {
      return outfd;
    }

  FILE *outfh = fdopen (outfd, "w");
  if (!outfh)
    {
      rc = OATH_FILE_CREATE_ERROR;
      goto out;
    }

  /* ownership has been transferred to outfh */
  outfd = -1;

  /* Create the new usersfile content. */
  rc = update_usersfile2 (username, otp, infh, outfh, lineptr, n,
			  timestamp, new_moving_factor, skipped_users);

  if (rc != OATH_OK)
      goto out;

  /* On success, flush the buffers. */
  if (fflush (outfh) != 0) {
    rc = OATH_FILE_FLUSH_ERROR;
    goto out;
  }

  /* On success, sync the disks. */
  if (fsync (fileno (outfh)) != 0) {
    rc = OATH_FILE_SYNC_ERROR;
    goto out;
  }

  /* On success, replace the usersfile with the new copy.
   * This does not follow symlinks in the target, the target will always be
   * replaced.
   * */
  if (renameat (ctx->parent_fd, newfilename, ctx->parent_fd, ctx->basename) != 0) {
    rc = OATH_FILE_RENAME_ERROR;
    goto out;
  }

  /* this name no longer exists now */
  newfilename[0] = '\0';

  /* make sure the directory is also synced such that directory inodes are written out */
  if (fsync(ctx->parent_fd) != 0) {
    rc = OATH_FILE_SYNC_ERROR;
    goto out;
  }

out:
  if (outfd >= 0)
      close(outfd);
  if (outfh)
      fclose(outfh);
  if (rc != OATH_OK && newfilename[0])
      unlinkat(ctx->parent_fd, newfilename, 0);
  return rc;
}

static int
oath_process_usersfile (struct usersfile_ctx *ctx,
			     const char *username,
			     const char *otp,
			     size_t window,
			     const char *passwd, time_t *last_otp)
{
  FILE *infh;
  char *line = NULL;
  size_t n = 0;
  uint64_t new_moving_factor;
  int rc;
  size_t skipped_users;

  infh = fdopen (ctx->fd, "r");
  if (infh == NULL)
    return OATH_FILE_OPEN_ERROR;

  /* ownership has been transferred to the FILE stream now */
  ctx->fd = -1;

  rc = parse_usersfile (username, otp, window, passwd, last_otp,
			infh, &line, &n, &new_moving_factor, &skipped_users);

  if (rc == OATH_OK)
    {
      char timestamp[30];
      size_t max = sizeof (timestamp);
      struct tm now;
      time_t t;
      size_t l;

      if (time (&t) == (time_t) - 1)
        return OATH_TIME_ERROR;

      if (localtime_r (&t, &now) == NULL)
        return OATH_TIME_ERROR;

      l = strftime (timestamp, max, TIME_FORMAT_STRING, &now);
      if (l != 20)
        return OATH_TIME_ERROR;

      rc = update_usersfile (ctx, username, otp, infh,
			     &line, &n, timestamp, new_moving_factor,
			     skipped_users);
    }

  free (line);
  fclose (infh);

  return rc;
}

/*
 * Safely open `ctx->path`, filling all the other fields in `ctx` from it. On
 * error destroy_usersfile_ctx() is invoked for `ctx`.
 *
 * When operating with raised privileges we cannot know the ownership of
 * `ctx->path` in advance, thus we need to carefully open the path. Any
 * symbolic links in the path will be rejected for simplicity reasons.
 *
 * Every path element will be extracted step-by-step and opened by passing the
 * `O_PATH` flag. This is the safest approach which prevents any side effects
 * that might result from opening e.g. FIFO special files, symlinks or device
 * files.
 *
 * Once the final path element has been reached and verified, the file
 * descriptors have to be upgraded to regular ones without the `O_PATH`
 * property, for being able to use them for regular file operations.
 *
 * NOTE: a similar result can be achieved by using openat2() and passing
 * RESOLVE_NO_SYMLINKS, but the system call is not yet wrapped in Glibc, which
 * makes it hard to use it.
 */
static int
safe_open_usersfile (struct usersfile_ctx *ctx)
{
  int err = OATH_OK;

  /* reject relative paths */
  if (ctx->path[0] != '/')
    return OATH_FILE_OPEN_ERROR;

  ctx->parent_fd = open("/", O_PATH|O_DIRECTORY|O_CLOEXEC|O_RDONLY);
  if (ctx->parent_fd < 0)
    return OATH_FILE_OPEN_ERROR;

  char *path_start = strdup (ctx->path);
  if (!path_start) {
      err = OATH_MALLOC_ERROR;
      goto out;
  }

  char *element = path_start;

  while (true)
    {
      /* ignore any extra leading slashes */
      while (element[0] == '/')
          element++;

      /* end of path has been reached (trailing slashes? shouldn't really happen) */
      if (!element[0])
        {
          err = OATH_FILE_OPEN_ERROR;
          goto out;
        }

      char *sep = strpbrk(element, "/");

      /* intermediate path (directory) element */
      if (sep)
        {
          *sep = '\0';

          ctx->fd = openat(ctx->parent_fd, element, O_RDONLY|O_PATH|O_CLOEXEC|O_NOFOLLOW|O_DIRECTORY);

          if (ctx->fd < 0)
            {
              err = errno == ENOENT ? OATH_NO_SUCH_FILE : OATH_FILE_OPEN_ERROR;
              goto out;
            }

          if (fstat(ctx->fd, &ctx->st) != 0)
            {
              err = OATH_FILE_STAT_ERROR;
              goto out;
            }

          /* If we encounter any world-writable components, refuse the path.
           * This prevents any unwise configurations like placing the file into
           * /var/tmp or a dedicated world-writable sticky-bit directory from
           * working. */
          if (ctx->st.st_mode & S_IWOTH)
            {
              err = OATH_FILE_OPEN_ERROR;
              goto out;
            }

          close(ctx->parent_fd);
          ctx->parent_fd = ctx->fd;
          ctx->fd = -1;
          element = sep + 1;
        }
      /* final path element has been encountered */
      else
        {
          ctx->basename = ctx->path + (element - path_start);
          err = finish_open_usersfile(ctx);
          break;
        }
    }


out:
  if (err != OATH_OK)
    {
      destroy_usersfile_ctx(ctx);
    }
  free (path_start);
  return err;
}

/**
 * oath_authenticate_usersfile:
 * @usersfile: string with user credential filename, in UsersFile format
 * @username: string with name of user
 * @otp: string with one-time password to authenticate
 * @window: how many past/future OTPs to search
 * @passwd: string with password, or NULL to disable password checking
 * @last_otp: output variable holding last successful authentication
 *
 * Authenticate user named @username with the one-time password @otp
 * and (optional) password @passwd.  Credentials are read (and
 * updated) from a text file named @usersfile.
 *
 * Note that for TOTP the usersfile will only record the last OTP and
 * use that to make sure more recent OTPs have not been seen yet when
 * validating a new OTP.  That logics relies on using the same search
 * window for the same user.
 *
 * Returns: On successful validation, %OATH_OK is returned.  If the
 *   supplied @otp is the same as the last successfully authenticated
 *   one-time password, %OATH_REPLAYED_OTP is returned and the
 *   timestamp of the last authentication is returned in @last_otp.
 *   If the one-time password is not found in the indicated search
 *   window, %OATH_INVALID_OTP is returned.  Otherwise, an error code
 *   is returned.
 **/
int
oath_authenticate_usersfile (const char *usersfile,
			     const char *username,
			     const char *otp,
			     size_t window,
			     const char *passwd, time_t *last_otp)
{
  int rc;
  struct usersfile_ctx ctx;
  init_usersfile_ctx(&ctx, usersfile);

  rc = safe_open_usersfile (&ctx);
  if (rc < 0)
    return rc;

  /* if user is not root we cannot change credentials,
     just run _oath_authenticate_usersfile normally in this case.
     Similarly if the file is owned by root, we don't need to change
     credentials. */
  if (geteuid () != 0 || ctx.st.st_uid == 0)
    {
        rc = oath_process_usersfile (&ctx, username, otp, window, passwd, last_otp);
        destroy_usersfile_ctx(&ctx);
        return rc;
    }

  /* else spawn a new process so we can drop privileges to the owner of the
   * file, to be on the safe side when operating in a directory owned by
   * non-root. */
  pid_t cpid = fork ();
  if (cpid < 0)
    {
        destroy_usersfile_ctx(&ctx);
        return OATH_FORK_ERROR;
    }

  if (cpid == 0)
    {
      /* child */
      if (setgroups(0, NULL) != 0)
        exit (abs(OATH_SETGROUPS_ERROR));
      if (setgid (ctx.st.st_gid) != 0)
        exit (abs(OATH_SETGID_ERROR));
      if (setuid (ctx.st.st_uid) != 0)
        exit (abs(OATH_SETUID_ERROR));
      rc = oath_process_usersfile (&ctx, username, otp, window, passwd, last_otp);
      exit (abs(rc));
    }
  else
    {
      int status;
      rc = waitpid (cpid, &status, 0);
      if (rc < 0)
        goto wait_out;

      if (!WIFEXITED(status))
        {
            /* child exited abnormally */
            rc = OATH_WAIT_ERROR;
            goto wait_out;
        }

      const int exit_code = WEXITSTATUS(status);
      rc = exit_code == 0 ? OATH_OK : -exit_code;
wait_out:
      /*
       * only destroy the ctx after the child exited, otherwise the lockfile
       * would be unlinked before the job is finished.
       */
      destroy_usersfile_ctx(&ctx);
      return rc;
    }
}

#else /* _WIN32 */

int
oath_authenticate_usersfile (const char *usersfile,
			     const char *username,
			     const char *otp,
			     size_t window,
			     const char *passwd, time_t *last_otp)
{
  (void) usersfile;
  (void) username;
  (void) otp;
  (void) window;
  (void) passwd;
  (void) last_otp;

  /* The main reason we don't support this on Windows yet is file
   * locking. So return that as the error. */
  return OATH_FILE_LOCK_ERROR;
}
#endif
